package main

import (
	"git.jfdev.de/JonasFranzDEV/hal/hal"
	"git.jfdev.de/JonasFranzDEV/hal/parser"
	"github.com/stretchr/testify/assert"
	"math"
	"testing"
)

func TestAddition(t *testing.T) {
	input := []string{
		"1 START",
		"2 IN 0",
		"3 STORE 10",
		"4 IN 0",
		"5 ADD 10",
		"6 OUT 1",
		"7 STOP",
	}

	program, err := parser.ParseProgram(input)
	assert.NoError(t, err)

	module, err := hal.NewHALModule(program, 256, 2, false)
	assert.NoError(t, err)

	module.IO[0] = 10

	err = module.Run()
	assert.NoError(t, err)

	assert.Equal(t, float64(20), module.IO[1])
}

func TestMax(t *testing.T) {
	input := []string{
		"00 IN 0",
		"01 STORE 1",
		"02 IN 1",
		"03 STORE 2",
		"04 SUB 1",
		"05 JUMPPOS 9",
		"06 LOAD 1",
		"07 OUT 1",
		"08 STOP",
		"09 LOAD 2",
		"10 OUT 1",
		"11 STOP",
	}

	program, err := parser.ParseProgram(input)
	assert.NoError(t, err)

	module, err := hal.NewHALModule(program, 256, 2, false)
	assert.NoError(t, err)

	module.IO[0] = 10
	module.IO[1] = 15

	err = module.Run()
	assert.NoError(t, err)

	assert.Equal(t, float64(15), module.IO[1])
}

func TestNewton1(t *testing.T) {
	input := []string{
		"01 START",
		"02 LOADNUM 1",
		"03 STORE 10",
		"10 MUL 10",
		"11 MUL 10",
		"12 STORE 12",
		"13 MUL 10",
		"14 MUL 10",
		"15 STORE 11",
		"21 LOAD 12",
		"22 MULNUM 5",
		"30 ADD 11",
		"31 SUBNUM 5",
		"32 STORE 20",
		"33 SUB 50",
		"34 JUMPNULL 66",
		"35 LOAD 20",
		"36 STORE 50",
		"40 LOAD 10",
		"41 MUL 10",
		"42 STORE 21",
		"43 MUL 10",
		"44 MUL 10",
		"45 MULNUM 5",
		"46 STORE 22",
		"47 LOAD 21",
		"48 MULNUM 15",
		"49 ADD 22",
		"50 STORE 30",
		"60 LOAD 20",
		"61 DIV 30",
		"62 STORE 40",
		"63 LOAD 10",
		"64 SUB 40",
		"65 JUMP 03",
		"66 LOAD 10",
		"67 OUT 1",
		"68 STOP",
	}

	program, err := parser.ParseProgram(input)
	assert.NoError(t, err)

	module, err := hal.NewHALModule(program, 256, 2, false)
	assert.NoError(t, err)

	err = module.Run()
	assert.NoError(t, err)

	assert.Equal(t, calculateNewton(1), module.IO[1])
}

func f1(x float64) float64 {
	return math.Pow(x, 5) + 5*math.Pow(x, 3) - 5
}

func f1derived(x float64) float64 {
	return 5*math.Pow(x, 4) + 15*math.Pow(x, 2)
}

func calculateNewton(start float64) float64 {
	var last float64
	x := start
	for x != last {
		last = x
		x = x - f1(x)/f1derived(x)
	}
	return x
}
